/*
 * $Id:$
 *
 * Copyright (c) 2015, Juniper Networks, Inc.
 * All rights reserved.
 *
 */

#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/queue.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#define __STDC_VERSION__ 199901L
#include <efi.h>

#ifdef BUILD_EFI
#include <efilib.h>
#else
#include <openssl/x509.h>
#include <openssl/bio.h>
#include <openssl/err.h>
#include <openssl/pem.h>
#endif


#include <guid.h>
#include <sha256.h>
#ifndef BUILD_EFI
#include <kernel_efivars.h>
#else
#include <variables.h>
#endif

#include "efiauthenticated.h"
#include "revocations.h"
#include "efi-jnpr.h"

#define ARRAY_SIZE(a) (sizeof (a) / sizeof ((a)[0]))

#ifndef BUILD_EFI
static int verbose;
#endif

#ifdef BUILD_EFI
#define fprintf(level, ...) do { } while (0);
#define printf(...) do { } while (0);
#define calloc(num, size) AllocateZeroPool((num) * (size))
#define malloc(size) AllocatePool((size))
#define memcmp(one, two, size) CompareMem((one), (two), (size))
#define memcpy(d, s, len) CopyMem((d), (s), (len))
#define free(ptr) FreePool((ptr))
#define verbose_printf(level, ...) do { } while (0)

#define set_variable(name, guid, attr, size, data) \
    uefi_call_wrapper(RT->SetVariable, 5, (name), (guid), (attr), (size), (data))

int
get_variable_alloc(void *var, EFI_GUID *guid, uint32_t *attributes,
		   uint32_t *size, UINT8 **buf)
{
	return get_variable_attr(var, buf, (UINTN *) size, *guid, attributes);

}
#else
#define verbose_printf(level, ...) do { if (verbose >= level) fprintf(stderr, __VA_ARGS__); } while (0);

void
inc_revocation_verbosity (int diff)
{
    verbose += diff;
}

void
set_revocation_verbosity (int level)
{
    verbose = level;
}
#endif

/* apologies for not too orignal Hexspeak */
EFI_GUID JSETUP_GUID = { 0xb105f00d, 0x1bad, 0xb001, {0xde, 0xad, 0xc0, 0xde, 0x00, 0x01, 0x02, 0x03 }};

/*
 * Convert efi_revocation_type_e to
 * string representation.
 */
const char *
efi_revocation_type_to_str(efi_revocation_type_e e)
{
	static const char *table[] = {
		"X509", /* EFI_REVOCATION_X509 */
		"SHA256", /* EFI_REVOCATION_SHA256 */
	}; /* this must be kept in sync with efi_revocation_type_e */

	if (e < 0 || e >= ARRAY_SIZE(table)) {
		return "<unknown>";
	} else {
		return table[e];
	}
}

#ifndef BUILD_EFI
/*
 * Dump out human readable representation
 * of EFI key
 */
static void
dump_revocation(efi_revocation_t *node)
{
	int i;

	printf("--------------------------------------------------------------------------------\n");
	if (node->ert == EFI_REVOCATION_X509) {
		const unsigned char *buf = (unsigned char *)node->value;
		X509 *X = d2i_X509(NULL, &buf,
		    node->value_len);
		X509_NAME *issuer = X509_get_issuer_name(X);
		X509_NAME *subject = X509_get_subject_name(X);

		printf("    X509 cert:\n");
		printf("        Subject:\n");
		X509_NAME_print_ex_fp(stdout, subject, 12, XN_FLAG_SEP_CPLUS_SPC);
		printf("\n        Issuer:\n");
		X509_NAME_print_ex_fp(stdout, issuer, 12, XN_FLAG_SEP_CPLUS_SPC);
		printf("\n");
		X509_free(X);

	} else if (node->ert == EFI_REVOCATION_SHA256) {
		uint8_t *hash = node->value;
		int j;

		printf("    SHA256 cert:\n");
		printf("        Hash:");
		for (j = 0; j < SHA256_DIGEST_SIZE; j++) {
			printf("%02x", hash[j]);
		}
		printf("\n");
	}
	verbose_printf(2, "type = %s\nval: ", efi_revocation_type_to_str(node->ert));
	for (i = 0; i < node->value_len; i++) {
		verbose_printf(2, "%02x", *((uint8_t *)(node->value) + i));
	}
	verbose_printf(2, "\n");
	printf("--------------------------------------------------------------------------------\n");

}
#endif

static struct revocation_set *
get_revocation_set(void)
{
	struct revocation_set *set;
	int status;
	uint32_t len;
	uint8_t *buf = NULL;

	set = NULL;

	status = get_variable_alloc(REVOKE_VAR_NAME, REVOKE_VAR_OWNER, NULL,
	    &len, &buf);

	if (status != 0) {
		fprintf(stderr, "Failed to get revocation list: %d\n", status);
		goto exit;
	} else {
		set = parse_revocations(buf, &len);
		if (set == NULL) {
		    fprintf(stderr, "ERROR: Couldn't parse revocation set\n");
		}
	}

exit:
	if (buf) {
		free(buf);
	}
	return set;
}

/*
 * Revoke all Juniper bringup keys we know about. Can be called
 * multiple times -- it does re-revoke anything.
 */
int
juniper_revoke_bringup_keys(void)
{
	int rc;
	struct revocation_set *set;
	set = get_revocation_set();

	verbose_printf(1, "Asserting status of bringup key revocation...\n");

	/*
	 * Revoke whatever else is supposed to be revoked.
	 */
	rc = efi_revoke_rest_bringup(set);

	verbose_printf(1, "Bringup revocation status: %s\n",
	    rc == 0 ? "Success" : "Failure");

	efi_revocation_set_free(set);
	set = NULL;

	return rc;

}

/*
 * Revoke all Juniper dev keys we know about. Can be called
 * multiple times -- it does not re-revoke anything.
 */
int
juniper_revoke_dev_keys(void)
{
	int rc;
	struct revocation_set *set;
	set = get_revocation_set();

	/* always attempt to revoke bringup keys */
	juniper_revoke_bringup_keys();

#ifndef BUILD_EFI
	efi_revocation_t *node;
	verbose_printf(1, "Attempting to revoke dev keys\n");

	if (verbose) {
		if (set == NULL || SLIST_EMPTY(set)) {
			verbose_printf(1, "There are no revoked keys in UEFI BIOS\n");
		} else {
			verbose_printf(1, "Currently revoked keys in UEFI BIOS:\n");
			SLIST_FOREACH(node, set, next)
				dump_revocation(node);
			verbose_printf(1, "End of currently revoked keys in UEFI BIOS\n");
		}
	}
#endif

	/*
	 * Revoke whatever else is supposed to be revoked.
	 */
	rc = efi_revoke_rest_dev(set);

	fprintf(stderr, "Dev key revocation status: %s\n",
		rc == 0 ? "Success" : "Failure");

	verbose_printf(1, "Attempting to leave revocation breadcrumb: ");
	set_revocation_breadcrumb();

	efi_revocation_set_free(set);
	set = NULL;

	verbose_printf(1, "All done\n");

	return rc;
}

static int
get_jsetup(uint64_t *val)
{
	EFI_GUID *owner = JSETUP_VAR_OWNER;
	uint8_t *data = NULL;
	uint32_t len;
	int rc;

	rc = get_variable_alloc(JSETUP_VAR_NAME, owner, NULL, &len, &data);
	if (rc != 0) {
	    verbose_printf(2, "JSetup var could not be fetched\n");
	    *val = 0;
	} else if (len != sizeof(uint64_t)) {
	    verbose_printf(2, "JSetup var: unexpected size\n");
	    *val = 0;
	} else {
	    *val = *(uint64_t *)data;
	    free(data);
	}


	verbose_printf(2, "JSetup found and fetched\n");
	return rc;
}

static int
set_jsetup(uint64_t val)
{
	EFI_GUID *owner = JSETUP_VAR_OWNER;
	uint32_t attributes = EFI_VARIABLE_NON_VOLATILE
	    | EFI_VARIABLE_RUNTIME_ACCESS
	    | EFI_VARIABLE_BOOTSERVICE_ACCESS;
	int rc;

	rc = set_variable(JSETUP_VAR_NAME, owner, attributes, sizeof(uint64_t), &val);

	return rc;
}

/*
 * Indicate that user has revoked Dev keys
 * via Jsetup variable
 */
int
set_revocation_breadcrumb(void)
{
	int rc = 0;
	uint64_t val;

	if (get_jsetup(&val) != 0) {
		fprintf(stderr, "Failed to fetch efi var: %s\n",
		    JSETUP_VAR_NAME);
		rc = -1;
		goto exit;
	}

	if ((val & JSETUP_DEV_KEYS_REVOKED) != 0) {
		verbose_printf(1, "Dev key breadcrumbs are already present\n");
		/* revocation already indicated */
		goto exit;
	}

	val |= JSETUP_DEV_KEYS_REVOKED;

	if (set_jsetup(val) != 0) {
		fprintf(stderr, "Failed to write to efi var: %s\n",
		    JSETUP_VAR_NAME);
		rc = -1;
		goto exit;
	}

	verbose_printf(1, "Success");

exit:
	return rc;
}

/*
 * See if bit is set in Jsetup variable
 */
static int
is_jsetup_bit_set(uint64_t bit)
{
	uint64_t val;

	if (get_jsetup(&val) != 0) {
		fprintf(stderr, "Failed to fetch efi var: %s\n",
		    JSETUP_VAR_NAME);
		return 0;
	}

	return ((val & bit) != 0);
}

/*
 * Consult with Jsetup to see if keys are supposed to be revoked
 */
static int
is_jsetup_dev_keys_revoked(void)
{
	return is_jsetup_bit_set(JSETUP_DEV_KEYS_REVOKED);
}

/*
 * Consult with Jsetup to see if secure boot defaults were restored
 */
static int
is_jsetup_secure_boot_defaults_restored(void)
{
	return is_jsetup_bit_set(JSETUP_FACTORY_SB_RESTORED);
}

/*
 * Return true if any of the dev keys we know about are revoked
 */
int
are_any_jnpr_dev_keys_revoked(void)
{
	int i, found = 0;
	struct revocation_set *set;
	efi_revocation_t *node;
	int offset, len;


	set = get_revocation_set();

	if (set == NULL)
		return -1;

	SLIST_FOREACH(node, set, next) {
		for (i = 0; i < ARRAY_SIZE(revocations); i++) {
			if (revocations[i].platform != platform) {
				verbose_printf(2, "Skipping check of %s."
				    " Platform mismatch (curplat=%d, targetplat=%d)\n",
				    revocations[i].id, platform,
				    revocations[i].platform);
				continue;
			}
			if (revocations[i].type != EFI_KEY_DEVELOPMENT) {
				verbose_printf(2, "Skipping check of %s."
				    " it's not key of type: %d\n",
				    revocations[i].id, EFI_KEY_DEVELOPMENT);
				continue;
			}
			if (node->value_len == revocations[i].payload_size) {
				len = revocations[i].value_end - revocations[i].value_start;
				offset = len - revocations[i].payload_size;

				if (offset < 0) {
					/* something's funny */
					continue;
				}

				if (memcmp(node->value,
				    &revocations[i].value_start[offset], node->value_len) == 0) {
					found += 1;
				}
			}
		}
	}

	efi_revocation_set_free(set);
	set = NULL;

	verbose_printf(2, "Found %d revoked dev keys for platform %d\n", found,
	    platform);
	return found;
}
/*
 *
 * Nothing to do; although in the future we may need
 * revoke keys (either due to compromise or something else)
 *
 * For now, we just consume the bit.
 */
static int
process_secure_boot_restoration(void)
{
	int rc = 0;
	uint64_t val;

	if (get_jsetup(&val) != 0) {
		fprintf(stderr, "Failed to fetch efi var: %s\n",
		    JSETUP_VAR_NAME);
		rc = -1;
		goto exit;
	}

	val &= ~JSETUP_FACTORY_SB_RESTORED;

	if (set_jsetup(val) != 0) {
		fprintf(stderr, "Failed to write to efi var: %s\n",
		    JSETUP_VAR_NAME);
		rc = -1;
		goto exit;
	}

	verbose_printf(0, "Successfully removed indication of"
		       " secure boot defaults restoration\n");

exit:
	return rc;

}

int
juniper_assert_revocation(void)
{
	int count, need_to_re_revoke = 0;

	/* always attempt to revoke bringup keys */
	juniper_revoke_bringup_keys();

	if (is_jsetup_dev_keys_revoked()) {
		need_to_re_revoke = 1;
		verbose_printf(1, "Accoring to %s, dev keys are already revoked\n",
		    JSETUP_VAR_NAME);
	} else if ((count = are_any_jnpr_dev_keys_revoked()) != 0) {
		need_to_re_revoke = 1;
		verbose_printf(1, "Found %d known dev revoked keys\n", count);
	} else if (is_jsetup_secure_boot_defaults_restored()) {
		verbose_printf(0, "According to %s, secure boot defaults"
			       " have been restored\n", JSETUP_VAR_NAME);
		process_secure_boot_restoration();
	}

	if (need_to_re_revoke) {
		verbose_printf(1, "Checking to see if more keys need to be revoked\n");
		return juniper_revoke_dev_keys();
	} else {
		/* Keys are not revoked, nothing to do */
		fprintf(stderr, "Junper Dev keys are not revoked. Doing nothing\n");
	}

	return 0;
}


/*
 * Return currently installed EFI dbx keys from dbx variable.
 * Keys are allocated from heap, and should be freed using
 * efi_revocation_set_free()
 */
struct revocation_set *
parse_revocations(uint8_t *buf, uint32_t *len)
{
	EFI_SIGNATURE_LIST  *CertList = (EFI_SIGNATURE_LIST *)buf;
	EFI_SIGNATURE_DATA  *Cert;
	long DataSize = *len;
	int size;
	struct revocation_set *set;
	efi_revocation_type_e type;
	efi_revocation_t *er;

	set = calloc(1, sizeof(*set));

	if (set == NULL) {
		fprintf(stderr, "Unable to allocate memory\n");
		return NULL;
	}

	if (CertList == NULL || len == 0) {
		/* empty set */
		return set;
	}

	certlist_for_each_certentry(CertList, buf, size, DataSize) {

		if (compare_guid(&CertList->SignatureType, &X509_GUID)== 0) {
			type = EFI_REVOCATION_X509;
		} else if (compare_guid(&CertList->SignatureType, &EFI_CERT_SHA256_GUID) == 0) {
			type = EFI_REVOCATION_SHA256;
		} else {
			fprintf(stderr, "Unexpected revocation type (%s)\n", "unknown");
			return NULL;
		}

		certentry_for_each_cert(Cert, CertList) {

			er = calloc(1, sizeof(*er));


			if (er == NULL) {
				fprintf(stderr, "Unable to allocate memory\n");
				return NULL;
			}
			er->ert = type;

			switch(type) {
			case EFI_REVOCATION_X509:
				er->value = calloc(1, CertList->SignatureSize -
				    sizeof(EFI_SIGNATURE_DATA) + 1);

				if (er->value == NULL) {
					fprintf(stderr, "Unable to allocate memory\n");
					return NULL;
				}
				er->value_len = CertList->SignatureSize -
				    sizeof(EFI_SIGNATURE_DATA) + 1;
				break;
			case EFI_REVOCATION_SHA256:
				er->value = calloc(1, SHA256_DIGEST_SIZE);

				if (er->value == NULL) {
					fprintf(stderr, "Unable to allocate memory\n");
					return NULL;
				}
				er->value_len = SHA256_DIGEST_SIZE;
				break;
			default:
				fprintf(stderr, "Unsupported key type (%d) detected\n", type);
				break;
			}

			if (er->value != NULL) {
				memcpy(er->value, Cert->SignatureData, er->value_len);
				SLIST_INSERT_HEAD(set, er, next);
			} else {
				/* Unknown type */
				free(er);
			}
		}
	}

	return set;
}

static void
efi_mark_already_revoked(efi_revocation_t *r)
{
	int i;
	int offset, len;

	for (i = 0; i < ARRAY_SIZE(revocations); i++) {
		if (r->value_len == revocations[i].payload_size) {
			len = revocations[i].value_end - revocations[i].value_start;
			offset = len - revocations[i].payload_size;

			if (offset < 0) {
				/* something's funny */
				continue;
			}

			if (memcmp(r->value, &revocations[i].value_start[offset], r->value_len) == 0) {
				revocations[i].revoked = 1;
			}
		}
	}
}

/*
 * Compares current bringup EFI keys with the ones that
 * are supposed to be revoked. Revokes the complement.
 */
int
efi_revoke_rest_bringup(struct revocation_set *set)
{
	return efi_revoke_rest(set, EFI_KEY_BRINGUP);
}

/*
 * Compares current dev EFI keys with the ones that
 * are supposed to be revoked. Revokes the complement.
 */
int
efi_revoke_rest_dev(struct revocation_set *set)
{
	return efi_revoke_rest(set, EFI_KEY_DEVELOPMENT);
}

/*
 * Compares current dev EFI keys with the ones that
 * are supposed to be revoked. Revokes the complement.
 */
int
efi_revoke_rest(struct revocation_set *set, efi_key_type_e type)
{
	efi_revocation_t *node;
	int i, ret, rc = 0;

	uint32_t attributes = EFI_VARIABLE_NON_VOLATILE
		| EFI_VARIABLE_RUNTIME_ACCESS
		| EFI_VARIABLE_BOOTSERVICE_ACCESS
		| EFI_VARIABLE_TIME_BASED_AUTHENTICATED_WRITE_ACCESS
		| EFI_VARIABLE_APPEND_WRITE;

	if (set) {
		SLIST_FOREACH(node, set, next) {
			efi_mark_already_revoked(node);
		}
	}

	for (i = 0; i < ARRAY_SIZE(revocations); i++) {
		if (revocations[i].revoked == 0) {

			if (revocations[i].type != type) {
				verbose_printf(2, "Skipping revocation of %s."
				    " it's not key of type: %d\n",
				    revocations[i].id, type);
				continue;
			}

			if (revocations[i].platform != platform) {
				verbose_printf(2, "Skipping revocation of %s."
				    " Platform mismatch (curplat=%d, targetplat=%d)\n",
				    revocations[i].id, platform,
				    revocations[i].platform);
				continue;
			}

			verbose_printf(1, "Attempting to revoke '%s': ", revocations[i].id);

			ret = set_variable(REVOKE_VAR_NAME, REVOKE_VAR_OWNER,
			    attributes, revocations[i].value_end - revocations[i].value_start,
			    revocations[i].value_start);

			verbose_printf(1, "%s ", ret == 0 ? "Success" : "Failure");

			if (ret == EACCES) {
				verbose_printf(1, "(Reason: unknown PK/KEK "
					       "installed and in User mode)\n");
			} else if (ret != 0) {
				verbose_printf(1, "(unkown reason %d)\n", ret);
			} else {
				verbose_printf(1, "\n");
			}
			rc |= ret;
		}
	}

	return rc;
}

/*
 * Free EFI keys stored in a revocation_set structure.
 */
void
efi_revocation_set_free(struct revocation_set *set)
{
	efi_revocation_t *er;

	if (set == NULL)
		return;

	while (!SLIST_EMPTY(set)) {
		er = SLIST_FIRST(set);
		SLIST_REMOVE_HEAD(set, next);
		free(er->value);
		free(er);
	}
	free(set);
}

