#!/bin/sh
#
# Copyright (c) 2015, Juniper Networks, Inc.
#
# Tool to assert validity of secure boot signatures. The following signature
# types are supported:
#
# EFI:
#    Assert validity of an EFI file under currently running UEFI BIOS.
#
#    Inputs:
#     - EFI binary to be verified
#     - EFIVARFS mounted file system
#    Outputs:
#     - echo OKAY; exit 0
#    or
#     - echo Error: $reason; exit -1
#
# GRUB:
#    Assert validity of an PSIG (GPG) siguature considering revocation via UEFI BIOS.
#
#    Inputs:
#     - ${Object}
#     - ${Object}.psig
#     - ${Pubkey}
#     - EFIVARFS mounted file system
#    Outputs:
#     - echo OKAY; exit 0
#    or
#     - echo Error: $reason; exit -1
#

Myname="$0"
Mydir=`dirname ${Myname}`

# Optional first argument for file type, defaults to -x.
# Second argument is the file or directory to be found.
# Third argument is the PATH to search.
Which() {
	case "$1" in
	-*) t=$1; shift ;;
	*) t=-x ;;
	esac
	case "$1" in
	/*) test $t $1 && echo $1 ;;
	*) for d in `IFS=:; echo ${2-$PATH}`
	do
		test $t $d/$1 && { echo $d/$1; break; }
	done
	;;
	esac
}

# temp hack until tools are properly integrated
export PATH=${PATH}:/home/root:/root

OPENSSL=${OPENSSL:-`Which openssl`}
SBVERIFY=${SBVERIFY:-`Which sbverify`}
READVAR=${READVAR:-`Which efi-readvar`}
GRUBVERIFY=${GRUBVERIFY:-`Which grub-verify`}
OD=${OD:-`Which od`}
WC=${WC:-`Which wc`}
SED=${SED:-`Which sed`}
MKTEMP=${MKTEMP:-`Which mktemp`}
CERTDIR=
verbose=:

Error() {
	echo "Error: $@" >&2
	exit 1
}

Notify() {
	echo "Notify: $@"
}

help() {
	Notify "$Myname [-v] [-h] [-e <EFI1> <EFI2> ... <EFI3>] [-g <object> <sig> <pubkey>]"
	exit 0
}

_verify_EFI_SHA256()
{
	# not supported, assume verification failure
	return 1
}

_verify_EFI_X509()
{
	local cert="$1" target="$2"

	${SBVERIFY} --cert ${cert} ${target} >/dev/null 2>&1
	return $?
}

_verify_EFI()
{
	local target="$1" allow ret tool pk_installed

	# failure by default
	allow=0

	# Is PK installed?
	# 0: false
	# >0: true
	pk_installed=0

	for db in ${CERTDIR}/db-*; do
		test -s "${db}" || continue
		tool=_verify_EFI_${db##*.}

		eval ${tool} ${db} ${target}
		case $? in
		0) # success
			$verbose Notify "${db##*/}(${target}) = success, db verification succeeded"
			allow=1
			;;
		1) # fail
			$verbose Notify "${db##*/}(${target}) = verification failure"
			;;
		esac
	done

	for PK in ${CERTDIR}/PK-*; do
		test -s "${PK}" || continue
		pk_installed=`expr $pk_installed + 1`
	done


	case ${allow} in
	1) 	# Check to see its not revoked
		$verbose Notify Checking if ${target} is revoked via dbx variable
		for dbx in ${CERTDIR}/dbx-*; do
			test -s "${dbx}" || continue
			tool=_verify_EFI_${dbx##*.}

			eval ${tool} ${dbx} ${target}
			case $? in
			0) # success
				$verbose Notify "${dbx##*/}(${target}) = success, revoked"
				allow=0
				;;
			1) # fail
				$verbose Notify "${dbx##*/}(${target}) = verification failure"
				;;
			esac

		done
		;;
	esac

	if [ $pk_installed -eq 0 ]; then
		# We're in setup mode. Secure boot is disabled
		allow=1
	fi

	case ${allow} in
	0) echo DENY:${target}; ret=1  ;;
	1) echo ALLOW:${target}; ret=0  ;;
	esac

	return ${ret}
}

verify_EFI()
{
	local ret=0

	for i in "$@"; do
		test -s "${i}" || {
		Notify EFI binary \"${i}\" is unspecified or is empty or doesnt exist
		help
		}
		_verify_EFI ${i}
		case $? in
		1) ret=1;;
		esac
	done

	return $ret
}

_verify_GRUB_psig()
{
	local target="$1" sig="$2" pubkey="$3"

	test -s "${target}" || Error "${target}" not found
	test -s "${sig}" || Error "${sig}" not found
	test -s "${pubkey}" || Error "${pubkey}" not found

	${GRUBVERIFY} -k "${pubkey}" -s "${sig}" -o "${target}"

	return $?
}

_list_GRUB_fingerprint()
{
	local pubkey="$1" line

	test -s "${pubkey}" || Error pubkey \"${pubkey}\" not found

	# pubkey may contain multiple subkeys -- print one out per line
	${GRUBVERIFY} -l -k "${pubkey}" | ${SED} -e 's,  .*,,g' -e 's, ,,g' | while read line
	do
		# These are just 160bit PGP fingerprints. They're prepended
		# with 96 bits of 0's to map to SHA256 hash value
		echo -n 000000000000000000000000
		echo ${line}
	done
}

verify_GRUB()
{
	local target="$1" sig="$2" pubkey="$3" allow ret

	test -f "${target}" || Error Object \"${target}\" does not exist
	test -f "${sig}" || Error Signature \"${sig}\" does not exist
	test -f "${pubkey}" || Error Pubkey \"${pubkey}\" does not exist

	# failure by default
	allow=0

	_verify_GRUB_psig ${target} ${sig} ${pubkey}

	case $? in
	0)
		allow=1
		$verbose Notify signature verification success
		;;
	*)
		allow=0
		$verbose Notify signature verification failure
		;;
	esac

	case ${allow} in
	1)
		for dbx in ${CERTDIR}/dbx-*SHA256; do
			test -s "${dbx}" || continue

			# Now, match dbx fingerprint with pubkey fingerprint

			# convert dbx cert to ascii hex
			have=`${OD} -vt x1 ${dbx} | \
				${SED} -e ':a;N;$!ba;s/[[:digit:]]\{3,\}//g;s/[\n ]//g'`

			# ensure dbx cert is properly decoded in to ascii hex

			for want in `_list_GRUB_fingerprint ${pubkey}`
			do

				test `echo -n "${want}" | ${WC} -c` -eq 64 || {
					$verbose Notify skipping ${want} pubkey
					continue
				}

				test ${want} = ${have} && {
					allow=0
					$verbose Notify ${dbx} is revoking ${pubkey}
				}
			done

		done
		;;
	*) # nothing to do, sig verification failed
		;;
	esac

	case ${allow} in
	0) echo DENY:${target}; ret=1  ;;
	1) echo ALLOW:${target}; ret=0  ;;
	esac

	return ${ret}
}

cleanup() {
	rm -rf ${CERTDIR}
}

doit() {
	local ret=

	# verify
	verify_${sigtype} "$@"
	ret=$?

	exit $ret
}

eval_precond_EFI()
{
	# Sanity check for inputs
	test -x "${OPENSSL}" || Error OpenSSL \"${OPENSSL}\" is not executable
	test -x "${SBVERIFY}" || Error Sbverify \"${SBVERIFY}\" is not executable
	test -x "${READVAR}" || Error Efi-readvar \"${READVAR}\" is not executable
	test -x "${SED}" || Error sed \"${SED}\" is not executable
	test -x "${MKTEMP}" || Error mktemp \"${MKTEMP}\" is not executable
}

eval_precond_GRUB()
{
	test -x "${READVAR}" || Error Efi-readvar \"${READVAR}\" is not executable
	test -x "${GRUBVERIFY}" || Error grub-verify \"${GRUBVERIFY}\" is not executable
	test -x "${OD}" || Error od \"${OD}\" is not executable
	test -x "${SED}" || Error sed \"${SED}\" is not executable
	test -x "${MKTEMP}" || Error mktemp \"${MKTEMP}\" is not executable
	test -x "${WC}" || Error wc \"${WC}\" is not executable
}

get_certs()
{
	# extract the certs
	CERTDIR=`(umask 077 && mktemp -d "${TMPDIR:-/tmp}/efivarXXXXXXX") 2>/dev/null` || \
		Error unable to create temporary dir

	trap "cleanup" 0 1 2 3 15


	for var in PK KEK db dbx
	do
		${READVAR} -v ${var} -d ${CERTDIR} >/dev/null 2>&1 || \
			Error ${READVAR} failed to read ${var} from UEFI bios
	done
}

# This drives type of verification. We currently support
# [-e] to verify EFI signatures
# [-g] to verify GRUB signatures
# default to efi
sigtype=EFI

while :
do
	case "$1" in
	*=*) eval "$1"; shift;;
	-e) sigtype=EFI; shift ;;
	-g) sigtype=GRUB; shift ;;
	-h) help; exit 0; shift;;
	-v) verbose=; shift;;
	-x) set -x; shift ;;
	*) break;;
	esac
done

eval_precond_${sigtype}
get_certs
doit $@
